/**

  Non-deterministic intervals

  This file contains functions to generate non-deterministic intervals, used
  to simulate multiple inputs. It is used to emulate Malardalen's ".ann"
  files containing annotations for multiple inputs.

  These functions are defined according to the current abstract domain and 
  precision of the analysis. Any modifications must be taken into account
  to avoid semantic inconsistencies.

  Example: an analysis that incorporates a simple type reconstruction might
  not be able to infer from a function such as 
    int nondet_int() { int n; return n; }
  that the return value contains indeed an integer value. It may return 'Top'
  and thus be unable to use its result.

  On the other hand, for a sophisticated analyzer, using a simple function
  such as
    int nondet_int() { int i; return i ? -2147483648 : 2147483647 }
  might not construct an interval containing every possible integer value,
  if the analyzer is smart enough to consider shifted or unsigned intervals.

*/

int nondet_int() {
  int i; return i ? -2147483648 : 2147483647;
}

int nondet_interval(int min, int max) {
  int i;
  int res = i ? -2147483648 : 2147483647;
  if (res < min) res = min;
  if (res > max) res = max;
  return res;
}

