(* Based on PrintRTL, but without pretty-printers and producing strings instead of 
   directly printing them. Not very efficient. *)

open Format
open Camlcoq
open Datatypes
open Maps
open Integers
open RTL
open PrintAST
open Op
open DumpOpIa32
open RtlDriverUtil

(* TODO: should it be a command-line option? *)
(* When activated, prints conditions using "pretty dump" (e.g. "x <s 2")
   instead of "raw dump" (e.g. "Ccompimm Clt x 2") *)
let option_pretty_dump_condition = true


let reg r : string =
  sprintf "x%ld" (camlint_of_positive r)

let rec regs = function
  | [] -> ""
  | [r] -> reg r
  | r1::rl -> sprintf "%s, %s" (reg r1) (regs rl)

let ros = function
  | Coq_inl r -> reg r
  | Coq_inr s -> sprintf "\"%s\"" (extern_atom s)

let dump_instruction i =
  match i with
    | Inop s ->
      let s = camlint_of_positive s in
      sprintf "goto %ld" s
    | Iop(op, args, res, s) ->
      sprintf "%s = %s"
        (reg res) (dump_operation reg (op, args))
    | Iload(chunk, addr, args, dst, s) ->
      sprintf "%s = %s[%s]"
        (reg dst) (name_of_chunk chunk)
        (dump_addressing reg (addr, args))
    | Istore(chunk, addr, args, src, s) ->
      sprintf "%s[%s] = %s"
        (name_of_chunk chunk)
        (dump_addressing reg (addr, args))
        (reg src)
    | Icall(sg, fn, args, res, s) ->
      sprintf "%s = %s(%s)"
        (reg res) (ros fn) (regs args)
    | Itailcall(sg, fn, args) ->
      sprintf "tailcall %s(%s)"
        (ros fn) (regs args)
    | Ibuiltin(ef, args, res, s) ->
      sprintf "%s = builtin %s(%s)"
        (reg res) (name_of_external ef) (regs args)
    | Icond(cond, args, s1, s2) ->
      sprintf "if (%s) goto %ld else goto %ld"
        (let pp = 
           if option_pretty_dump_condition then 
             pretty_dump_condition 
           else dump_condition in
         pp reg (cond, args))
        (camlint_of_positive s1) (camlint_of_positive s2)
    | Ijumptable(arg, tbl) ->
      (*let tbl = Array.of_list tbl in*)
      sprintf "jumptable (%s) { <TODO> }" (reg arg);
(*      for i = 0 to Array.length tbl - 1 do
        sprintf "  case %d: goto %ld | " i (camlint_of_positive tbl.(i))
      done;
      sprintf "}"*)
    | Ireturn None ->
      sprintf "return"
    | Ireturn (Some arg) ->
      sprintf "return %s" (reg arg)

let dump_node f n : string =
  dump_instruction (do_opt (PTree.get n f.fn_code))
