open Printf
open Camlcoq
open Maps
open IntervalDomain
open VIntDomain
    
let itv x =
  let min = camlint64_of_z x.Interval.min in
  let max = camlint64_of_z x.Interval.max in
    sprintf "[%Ld; %Ld]" min max
      
let value (i,vi) =
  match vi with
    | MustVInt.Top -> "??"
    | MustVInt.VI -> itv i
    
let ptree used_vars f t =
  (* We sort it to help debugging *)
  let compare p1 p2 = (Int32.to_int (Camlcoq.camlint_of_positive (fst p1))) - (Int32.to_int (Camlcoq.camlint_of_positive (fst p2))) in
  (* Here, to print the information about all relevant variables (even those
     which have not been bounded and therefore are not printed, I use an
     ugly and inefficient hack... *)
  let seen_vars = ref [] in
  let s = List.fold_left
    (fun s (pos, v) -> sprintf "%s %sx%ld:%s" s (if (List.mem pos used_vars) then (seen_vars := pos :: !seen_vars; "_") else "") (camlint_of_positive pos) (f v))
    "" (List.sort compare (PTree.elements t)) (* we could use PTree.fold here,
                                                 but then we would not be able
                                                 to choose the iteration order *)
  in
  let s = s ^ "  " in (* Extra space to help identifying unseen variables *)
  let not_seen_vars = List.filter (fun x -> not (List.mem x !seen_vars)) used_vars in
  (* The main objective is to be sure no relevant variable has been forgotten,
     to understand the results obtained in the analysis. *)
  List.fold_left (fun s0 x -> sprintf "%s _x%s" s0 (SliceToString.binPosToString x)) s not_seen_vars

let regs used_vars itv =
  match itv with
    | None -> "BOT"
    | Some t -> ptree used_vars value t

(* Version used during "standard" printing (without loop bounds) *)
let analyze_and_show ?rel_vars:(xs=[]) ?slice_crit:(nc=[]) f = 
  let analyze = ValueAnalysis.analyze f in
    (fun n -> regs (if List.mem (camlint_of_positive n) nc then xs else []) (analyze n))

(* Version used when loop bounds are computed (to avoid recomputing the value analysis *)
let show dom_regs ?rel_vars:(xs=[]) ?slice_crit:(nc=[]) f = 
  (fun n -> regs (if List.mem (camlint_of_positive n) nc then xs else []) (dom_regs n))
