open Printf
open Camlcoq
open Errors
open Compiler

let print_error oc msg =
  let print_one_error = function
  | Errors.MSG s -> output_string oc (Camlcoq.camlstring_of_coqstring s)
  | Errors.CTX i -> output_string oc (Camlcoq.extern_atom i)
  | Errors.CTXL i -> output_string oc (extern_atom i)
  in
    List.iter print_one_error msg;
    output_char oc '\n'

let do_res (r : 'a res) : 'a = 
  match r with
    | OK v -> v
    | Error msg ->
      print_error stderr msg;
      exit 2

let do_opt (o : 'a option) : 'a = 
  match o with
    | Some a -> a
    | None ->
      fprintf stderr "do_opt: None\n";
      exit 2

let get_main (rtl_prog : RTL.program) : (AST.ident * RTL.coq_function) =
  let main_id = rtl_prog.AST.prog_main in
  let main_fd = List.assoc main_id rtl_prog.AST.prog_funct in
  match main_fd with
    | AST.External _ -> failwith "Assertion error: main is external"
    | AST.Internal main_f -> (main_id, main_f)
