open Printf

open Errors
open Datatypes
open RTL
open Peano
open Maps
open BinPos
open Op
open Registers
open Kildall
open Lattice
open Registers
open VarsUseDef
open List
open TheoryList

open ToString


(** Relevant variables computation **)

type varset = Regset.t
let var_to_reg r = coq_Ppred r
let reg_to_var r = coq_Psucc r
let memvar = Coq_xH


module L = LFSet(Regset)

let kill_no_mem rs kill =
  Regset.diff rs (Regset.remove Coq_xH kill)

(** val transfer : coq_function -> node list -> node -> L.t -> L.t **)

let transfer f slice_nodes pc rs =
  let kill = def f pc in
  let killed_rs = kill_no_mem rs kill in
  if mem positive_eq_dec pc slice_nodes
  then L.lub killed_rs (use f pc)
  else killed_rs

module Solver = Backward_Dataflow_Solver(L)(NodeSetBackward)

(** val compute_relevant_vars :
    coq_function -> node list -> L.t PMap.t option **)

let compute_relevant_vars f slice_nodes =
  Solver.fixpoint (successors f) (transfer f slice_nodes)
    (map (fun n -> (n, (use f n))) slice_nodes)

(** val compute_RV_in : coq_function -> node list -> L.t PMap.t **)
let compute_RV_in f slice_nodes : L.t PMap.t =
  match compute_relevant_vars f slice_nodes with
  | Some rvo ->
    ((fst rvo),
      (PTree.map (fun p lat -> transfer f slice_nodes p lat) (snd rvo)))
  | None -> (Regset.empty, PTree.empty)

(** Function defined as a Parameter in Coq; its result is validated by a checked. **)
let f_rvs (f : coq_function) (slice_nodes : node list) : node -> varset =
  let rvin = compute_RV_in f slice_nodes in
  Clflags.debugp "rel_var" (fun f -> f "%s\n" (pMapToString (fun s -> listToString binPosToString (Regset.elements s)) rvin));
  (fun n -> PMap.get n rvin)
