open Camlcoq
open AST
open BinInt
open BinPos
open Datatypes
open Globalenvs
open Maps
open Memory
open Op
open Registers
open Values
open RTL
open BourdoncleIterator

let int_of_positive p =
  let i = camlint_of_positive p in
    (Int32.to_int i) -1
      
let rec positive_of_int n =
  if n = 0 then assert false else
  if n = 1 then Coq_xH else
  if n land 1 = 0
  then Coq_xO (positive_of_int (n lsr 1))
  else Coq_xI (positive_of_int (n lsr 1))
let positive_of_int n = positive_of_int (n+1)

let iterator order_dec join widen top bot (transfer : (instruction -> (node * ('a -> 'a)) list)) f init =
  let cfg =
    let entry = int_of_positive f.fn_entrypoint in
    let max = PTree.fold (fun m n _ -> max m (int_of_positive n)) f.fn_code 0 in
    let succ = Array.make (max+1) [] in
    let _ = PTree.fold (fun () n ins ->
			  let succs = transfer ins in
			  let succs = List.map (fun (n',tf) -> (int_of_positive n',tf)) succs in
			    succ.(int_of_positive n) <- succs)
	      f.fn_code () in
      { entry = entry; succ = succ } in
  let adom = {
    order = order_dec;
    join = join;
    widen = widen;
    top = (fun _ -> top);
    bot = (fun _ -> bot);
    transfer = (fun f -> f);
    to_string = (fun _ -> "")
  } in
  let b = get_bourdoncle cfg in
  let run_with_bourdoncle = BourdoncleIterator.run_with_bourdoncle adom in
  let res = run_with_bourdoncle b cfg in
  let res_coq = ref (PMap.init bot) in      
    PTree.fold (fun _ p _ ->
		  res_coq := PMap.set p res.(int_of_positive p) !res_coq)
      (fn_code f) ();
    !res_coq
